@@ -6,7 +6,6 @@ use seq_macro::seq;
 use std::arch::x86::*;
 #[cfg(target_arch = "x86_64")]
 use std::arch::x86_64::*;
-use std::mem;
 
 trait NeedleWithSize: Needle {
     #[inline]
@@ -26,40 +25,15 @@ trait NeedleWithSize: Needle {
 
 impl<N: Needle + ?Sized> NeedleWithSize for N {}
 
-/// Rolling hash for the simple Rabin-Karp implementation. As a hashing
-/// function, the sum of all the bytes is computed.
-#[derive(Clone, Copy, Default, PartialEq)]
-struct ScalarHash(usize);
-
-impl From<&[u8]> for ScalarHash {
-    #[inline]
-    fn from(bytes: &[u8]) -> Self {
-        bytes.iter().fold(Default::default(), |mut hash, &b| {
-            hash.push(b);
-            hash
-        })
-    }
-}
-
-impl ScalarHash {
-    #[inline]
-    fn push(&mut self, b: u8) {
-        self.0 = self.0.wrapping_add(b.into());
-    }
-
-    #[inline]
-    fn pop(&mut self, b: u8) {
-        self.0 = self.0.wrapping_sub(b.into());
-    }
-}
-
 /// Represents an SIMD register type that is x86-specific (but could be used
 /// more generically) in order to share functionality between SSE2, AVX2 and
 /// possibly future implementations.
 trait Vector: Copy {
+    const LANES: usize;
+
     unsafe fn set1_epi8(a: i8) -> Self;
 
-    unsafe fn loadu_si(a: *const Self) -> Self;
+    unsafe fn loadu_si(a: *const u8) -> Self;
 
     unsafe fn cmpeq_epi8(a: Self, b: Self) -> Self;
 
@@ -68,7 +42,126 @@ trait Vector: Copy {
     unsafe fn movemask_epi8(a: Self) -> i32;
 }
 
+#[derive(Clone, Copy)]
+#[repr(transparent)]
+#[allow(non_camel_case_types)]
+struct __m16i(__m128i);
+
+impl Vector for __m16i {
+    const LANES: usize = 2;
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn set1_epi8(a: i8) -> Self {
+        __m16i(_mm_set1_epi8(a))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn loadu_si(a: *const u8) -> Self {
+        __m16i(_mm_set1_epi16(std::ptr::read_unaligned(a as *const i16)))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn cmpeq_epi8(a: Self, b: Self) -> Self {
+        __m16i(_mm_cmpeq_epi8(a.0, b.0))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn and_si(a: Self, b: Self) -> Self {
+        __m16i(_mm_and_si128(a.0, b.0))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn movemask_epi8(a: Self) -> i32 {
+        _mm_movemask_epi8(a.0) & 0x3
+    }
+}
+
+#[derive(Clone, Copy)]
+#[repr(transparent)]
+#[allow(non_camel_case_types)]
+struct __m32i(__m128i);
+
+impl Vector for __m32i {
+    const LANES: usize = 4;
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn set1_epi8(a: i8) -> Self {
+        __m32i(_mm_set1_epi8(a))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn loadu_si(a: *const u8) -> Self {
+        __m32i(_mm_set1_epi32(std::ptr::read_unaligned(a as *const i32)))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn cmpeq_epi8(a: Self, b: Self) -> Self {
+        __m32i(_mm_cmpeq_epi8(a.0, b.0))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn and_si(a: Self, b: Self) -> Self {
+        __m32i(_mm_and_si128(a.0, b.0))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn movemask_epi8(a: Self) -> i32 {
+        _mm_movemask_epi8(a.0) & 0xF
+    }
+}
+
+#[derive(Clone, Copy)]
+#[repr(transparent)]
+#[allow(non_camel_case_types)]
+struct __m64i(__m128i);
+
+impl Vector for __m64i {
+    const LANES: usize = 8;
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn set1_epi8(a: i8) -> Self {
+        __m64i(_mm_set1_epi8(a))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn loadu_si(a: *const u8) -> Self {
+        __m64i(_mm_set1_epi64x(std::ptr::read_unaligned(a as *const i64)))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn cmpeq_epi8(a: Self, b: Self) -> Self {
+        __m64i(_mm_cmpeq_epi8(a.0, b.0))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn and_si(a: Self, b: Self) -> Self {
+        __m64i(_mm_and_si128(a.0, b.0))
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn movemask_epi8(a: Self) -> i32 {
+        _mm_movemask_epi8(a.0) & 0xFF
+    }
+}
+
 impl Vector for __m128i {
+    const LANES: usize = 16;
+
     #[inline]
     #[target_feature(enable = "avx2")]
     unsafe fn set1_epi8(a: i8) -> Self {
@@ -77,8 +170,8 @@ impl Vector for __m128i {
 
     #[inline]
     #[target_feature(enable = "avx2")]
-    unsafe fn loadu_si(a: *const Self) -> Self {
-        _mm_loadu_si128(a)
+    unsafe fn loadu_si(a: *const u8) -> Self {
+        _mm_loadu_si128(a as *const Self)
     }
 
     #[inline]
@@ -101,6 +194,8 @@ impl Vector for __m128i {
 }
 
 impl Vector for __m256i {
+    const LANES: usize = 32;
+
     #[inline]
     #[target_feature(enable = "avx2")]
     unsafe fn set1_epi8(a: i8) -> Self {
@@ -109,8 +204,8 @@ impl Vector for __m256i {
 
     #[inline]
     #[target_feature(enable = "avx2")]
-    unsafe fn loadu_si(a: *const Self) -> Self {
-        _mm256_loadu_si256(a)
+    unsafe fn loadu_si(a: *const u8) -> Self {
+        _mm256_loadu_si256(a as *const Self)
     }
 
     #[inline]
@@ -151,6 +246,36 @@ impl<V: Vector> VectorHash<V> {
     }
 }
 
+impl From<&VectorHash<__m128i>> for VectorHash<__m64i> {
+    #[inline]
+    fn from(hash: &VectorHash<__m128i>) -> Self {
+        Self {
+            first: __m64i(hash.first),
+            last: __m64i(hash.last),
+        }
+    }
+}
+
+impl From<&VectorHash<__m128i>> for VectorHash<__m32i> {
+    #[inline]
+    fn from(hash: &VectorHash<__m128i>) -> Self {
+        Self {
+            first: __m32i(hash.first),
+            last: __m32i(hash.last),
+        }
+    }
+}
+
+impl From<&VectorHash<__m128i>> for VectorHash<__m16i> {
+    #[inline]
+    fn from(hash: &VectorHash<__m128i>) -> Self {
+        Self {
+            first: __m16i(hash.first),
+            last: __m16i(hash.last),
+        }
+    }
+}
+
 /// Single-substring searcher using an AVX2 algorithm based on the "Generic
 /// SIMD" algorithm [presented by Wojciech
 /// Muła](http://0x80.pl/articles/simd-strfind.html).
@@ -182,7 +307,6 @@ impl<V: Vector> VectorHash<V> {
 /// Rabin-Karp implementation.
 pub struct Avx2Searcher<N: Needle> {
     position: usize,
-    scalar_hash: ScalarHash,
     sse2_hash: VectorHash<__m128i>,
     avx2_hash: VectorHash<__m256i>,
     needle: N,
@@ -222,41 +346,17 @@ impl<N: Needle> Avx2Searcher<N> {
             assert_eq!(size, bytes.len());
         }
 
-        let scalar_hash = ScalarHash::from(bytes);
         let sse2_hash = VectorHash::new(bytes[0], bytes[position]);
         let avx2_hash = VectorHash::new(bytes[0], bytes[position]);
 
         Self {
-            needle,
             position,
-            scalar_hash,
             sse2_hash,
             avx2_hash,
+            needle,
         }
     }
 
-    #[inline]
-    fn scalar_search_in(&self, haystack: &[u8]) -> bool {
-        debug_assert!(haystack.len() >= self.needle.size());
-
-        let mut end = self.needle.size() - 1;
-        let mut hash = ScalarHash::from(&haystack[..end]);
-
-        while end < haystack.len() {
-            hash.push(*unsafe { haystack.get_unchecked(end) });
-            end += 1;
-
-            let start = end - self.needle.size();
-            if hash == self.scalar_hash && haystack[start..end] == *self.needle.as_bytes() {
-                return true;
-            }
-
-            hash.pop(*unsafe { haystack.get_unchecked(start) });
-        }
-
-        false
-    }
-
     #[inline]
     #[target_feature(enable = "avx2")]
     unsafe fn vector_search_in_chunk<V: Vector>(
@@ -266,8 +366,8 @@ impl<N: Needle> Avx2Searcher<N> {
         start: *const u8,
         mask: i32,
     ) -> bool {
-        let first = Vector::loadu_si(start.cast());
-        let last = Vector::loadu_si(start.add(self.position).cast());
+        let first = Vector::loadu_si(start);
+        let last = Vector::loadu_si(start.add(self.position));
 
         let eq_first = Vector::cmpeq_epi8(hash.first, first);
         let eq_last = Vector::cmpeq_epi8(hash.last, last);
@@ -316,19 +416,12 @@ impl<N: Needle> Avx2Searcher<N> {
     unsafe fn vector_search_in<V: Vector>(
         &self,
         haystack: &[u8],
+        end: usize,
         hash: &VectorHash<V>,
-        next: unsafe fn(&Self, &[u8]) -> bool,
     ) -> bool {
         debug_assert!(haystack.len() >= self.needle.size());
 
-        let lanes = mem::size_of::<V>();
-        let end = haystack.len() - self.needle.size() + 1;
-
-        if end < lanes {
-            return next(self, haystack);
-        }
-
-        let mut chunks = haystack[..end].chunks_exact(lanes);
+        let mut chunks = haystack[..end].chunks_exact(V::LANES);
         while let Some(chunk) = chunks.next() {
             if self.vector_search_in_chunk(haystack, hash, chunk.as_ptr(), -1) {
                 return true;
@@ -337,8 +430,8 @@ impl<N: Needle> Avx2Searcher<N> {
 
         let remainder = chunks.remainder().len();
         if remainder > 0 {
-            let start = haystack.as_ptr().add(end - lanes);
-            let mask = -1 << (lanes - remainder);
+            let start = haystack.as_ptr().add(end - V::LANES);
+            let mask = -1 << (V::LANES - remainder);
 
             if self.vector_search_in_chunk(haystack, hash, start, mask) {
                 return true;
@@ -350,25 +443,60 @@ impl<N: Needle> Avx2Searcher<N> {
 
     #[inline]
     #[target_feature(enable = "avx2")]
-    unsafe fn sse2_search_in(&self, haystack: &[u8]) -> bool {
-        self.vector_search_in(haystack, &self.sse2_hash, Self::scalar_search_in)
+    unsafe fn sse2_2_search_in(&self, haystack: &[u8], end: usize) -> bool {
+        let hash = VectorHash::<__m16i>::from(&self.sse2_hash);
+        self.vector_search_in(haystack, end, &hash)
     }
 
     #[inline]
     #[target_feature(enable = "avx2")]
-    unsafe fn avx2_search_in(&self, haystack: &[u8]) -> bool {
-        self.vector_search_in(haystack, &self.avx2_hash, Self::sse2_search_in)
+    unsafe fn sse2_4_search_in(&self, haystack: &[u8], end: usize) -> bool {
+        let hash = VectorHash::<__m32i>::from(&self.sse2_hash);
+        self.vector_search_in(haystack, end, &hash)
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn sse2_8_search_in(&self, haystack: &[u8], end: usize) -> bool {
+        let hash = VectorHash::<__m64i>::from(&self.sse2_hash);
+        self.vector_search_in(haystack, end, &hash)
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn sse2_16_search_in(&self, haystack: &[u8], end: usize) -> bool {
+        self.vector_search_in(haystack, end, &self.sse2_hash)
+    }
+
+    #[inline]
+    #[target_feature(enable = "avx2")]
+    unsafe fn avx2_search_in(&self, haystack: &[u8], end: usize) -> bool {
+        self.vector_search_in(haystack, end, &self.avx2_hash)
     }
 
     /// Inlined version of `search_in` for hot call sites.
     #[inline]
     #[target_feature(enable = "avx2")]
     pub unsafe fn inlined_search_in(&self, haystack: &[u8]) -> bool {
-        if haystack.len() < self.needle.size() {
-            return false;
+        if haystack.len() <= self.needle.size() {
+            return haystack == self.needle.as_bytes();
         }
 
-        self.avx2_search_in(haystack)
+        let end = haystack.len() - self.needle.size() + 1;
+
+        if end < __m16i::LANES {
+            unreachable!();
+        } else if end < __m32i::LANES {
+            self.sse2_2_search_in(haystack, end)
+        } else if end < __m64i::LANES {
+            self.sse2_4_search_in(haystack, end)
+        } else if end < __m128i::LANES {
+            self.sse2_8_search_in(haystack, end)
+        } else if end < __m256i::LANES {
+            self.sse2_16_search_in(haystack, end)
+        } else {
+            self.avx2_search_in(haystack, end)
+        }
     }
 
     /// Performs a substring search for the `needle` within `haystack`.
